unit UthDeskIcon;

interface

uses
  Classes, Windows;

const

  WORMSLIFECYCLE = 300;
  WORMLIFECYCLE  = WORMSLIFECYCLE-20;
  ATOMSLIFECYCLE = 300;
  MOUSESLIFECYCLE= 3000;
type

  TProcessEvent = function (AStoped : boolean) : integer of object;

  TItemInfo = class(TObject)
  private
    FX  : extended;
    FY  : extended;
    FDX : extended;
    FDY : extended;
    FCDX : extended;
    FCDY : extended;
    FR  : extended;
    FIX : integer;
    FIY : integer;
    FIndex : integer;
    FStep  : integer;
    FOwner : TObject;
    FLevel : integer;
    FLinks : TList;
  public
    constructor Create;
    destructor Destroy; override;
  end;

  TWormInfo = class(TObject)
  private
    FItems    : TList;
    FStep     : integer;
    FSkip     : integer;
    FLevel    : integer;
    FNextStep : boolean;
    FDX,FDY   : integer;
    FNX,FNY   : integer;
    FDirection: integer;
  public
    constructor Create;
    destructor Destroy; override;
  end;


  DeskIconThread = class(TThread)
  private
    HDeskList : THandle;
    FOnProcess : TProcessEvent;
    FItems : TList;
    FWorms : TList;
    FViewPort : TRect;
    FItemSpacing : integer;
    FWormStep : integer;
    FAtomStep : integer;
    FVDX,FVDY : integer;
    FMouseP : TPoint;
    { Private declarations }
  protected
    function  ProcessWait(AStoped : boolean) : integer;

    function  InitWorms(AStoped: boolean): integer;
    function  ProcessWorms(AStoped: boolean): integer;
    function  NextStepWorm(AWorm : TWormInfo): integer;
    function  FreeWorms(AStoped: boolean): integer;

    function  InitAtoms(AStoped: boolean): integer;
    function  ProcessAtoms(AStoped: boolean): integer;
    function  FreeAtoms(AStoped: boolean): integer;

    function  InitMouse(AStoped: boolean): integer;
    function  ProcessMouse(AStoped: boolean): integer;
    function  FreeMouse(AStoped: boolean): integer;

    function  ProcessSleep : integer;
    procedure ProcessDeskList(AHandle:THandle);
    procedure Execute; override;
  end;

implementation

uses DeskHelp, CommCtrl, Math;

{ DeskIconThread }

procedure DeskIconThread.Execute;
begin
  FOnProcess:=ProcessWait;
  sleep(100000);
  while not Terminated do begin
    sleep(ProcessSleep);
  end;
end;

procedure DeskIconThread.ProcessDeskList(AHandle: THandle);
begin
  ListView_SetItemPosition(AHandle,random(10),random(400),random(400));
end;

function DeskIconThread.ProcessSleep: integer;
begin
  try
    HDeskList:=GetDesktopListViewHandle;
    result:=FOnProcess(HDeskList=0);
  except
    result:=1000;
    FOnProcess:=ProcessWait;
  end;
end;

function DeskIconThread.ProcessWait(AStoped: boolean): integer;
begin
  result:=1000;
  if AStoped then exit;
  case random(105) of
    10..20 : FOnProcess:=InitWorms;
    25..35 : FOnProcess:=InitMouse;
    40..50 : FOnProcess:=InitAtoms;
    55..65 : FOnProcess:=InitWorms;
    70..80 : FOnProcess:=InitMouse;
    85..95 : FOnProcess:=InitAtoms;
  end;
end;

function DeskIconThread.InitWorms(AStoped: boolean): integer;
var
  p : TPoint;
  i,c : integer;
  LItem : TItemInfo;
  LWorm : TWormInfo;
  LItemRect : TRect;
begin
  result:=10;
  FOnProcess:=ProcessWait;
  if AStoped then exit;

  c:=ListView_GetItemCount(HDeskList);
  if c=0 then exit;
  if not ListView_GetViewRect(HDeskList,FViewPort) then exit;
  if not ListView_GetItemRect(HDeskList,0,LItemRect,LVIR_BOUNDS) then exit;

  FItemSpacing:=((LItemRect.Right-LItemRect.Left)+(LItemRect.Bottom-LItemRect.Top)) div 2;

  FVDX:=(FViewPort.Right-FViewPort.Left-FItemSpacing div 2) div FItemSpacing;
  FVDY:=(FViewPort.Bottom-FViewPort.Top-FItemSpacing div 2) div FItemSpacing;

  FItems := TList.Create;
  FWorms := TList.Create;
  FOnProcess:=ProcessWorms;
  FWormStep:=0;

  for i:=0 to c-1 do begin
    if ListView_GetItemPosition(HDeskList, i, p) then begin
      LItem := TItemInfo.Create;
      LItem.FIX:= p.x div FItemSpacing;
      LItem.FIY:= p.y div FItemSpacing;
      LItem.FX:= LItem.FIX*FItemSpacing;
      LItem.FY:= LItem.FIY*FItemSpacing;
      LItem.FIndex:=i;
      LItem.FLevel:=1;
      LItem.FStep:=FWormStep;
      FItems.Add(LItem);

      LWorm := TWormInfo.Create;
      LWorm.FItems.Add(LItem);
      LItem.FOwner:=LWorm;
      FWorms.Add(LWorm);
    end else break;
  end;
end;

function DeskIconThread.ProcessWorms(AStoped: boolean): integer;
var
  i : integer;
  LItem : TItemInfo;
begin
  result:=100;
  FOnProcess:=FreeWorms;
  if AStoped then exit;
  Inc(FWormStep);

  if (FWormStep<WORMSLIFECYCLE) then FOnProcess:=ProcessWorms;
  if (FWorms.Count=1) and (FWormStep<WORMLIFECYCLE) then FWormStep:=WORMLIFECYCLE;

  for i:=0 to FItems.Count-1 do begin
    LItem:=TItemInfo(FItems[i]);
    if LItem.FStep<FWormStep then result:=NextStepWorm(TWormInfo(LItem.FOwner));
    if LItem.FR=1 then
      ListView_SetItemPosition(HDeskList,LItem.FIndex,trunc(LItem.FX),trunc(LItem.FY));
  end;

end;

function DeskIconThread.FreeWorms(AStoped: boolean): integer;
var
  i : integer;
begin
  result:=100;
  FOnProcess:=ProcessWait;
  for i:=0 to FItems.Count-1 do TItemInfo(FItems[i]).Free;
  for i:=0 to FWorms.Count-1 do TWormInfo(FWorms[i]).Free;
  FItems.Free;
  FWorms.Free;
end;

function DeskIconThread.NextStepWorm(AWorm: TWormInfo): integer;
var
  i,c,t : integer;
  LItem,IItem : TItemInfo;
  LWorm: TWormInfo;
begin
  result:=100;

  with AWorm do begin

    LItem:=TItemInfo(AWorm.FItems[0]);

    if FStep=0 then begin
      if FSkip>5 then begin
        FSkip:=0;
        FLevel:=0-FLevel;
      end;
      FDirection:=random(101) mod 4;
      FStep:=random(10)+1;
    end;

    FDX:=0;
    FDY:=0;
    FNX:=0;
    FNY:=0;

    case FDirection of
      0:FDX:=1;
      1:FDY:=1;
      2:FDX:=-1;
      3:FDY:=-1;
    end;

    FNX:=LItem.FIX+FDX;
    FNY:=LItem.FIY+FDY;

    FNextStep:=true;

    for i:=0 to Self.FItems.Count-1 do begin
      IItem:=TItemInfo(Self.FItems[i]);
      if (IItem.FLevel=FLevel) then
        if (IItem.FIX=FNX) and (IItem.FIY=FNY) then
          if IItem.FOwner=AWorm  then begin
            FNextStep:=false;
            break;
          end else begin
            LWorm:=TWormInfo(IItem.FOwner);
            c:=LWorm.FItems.IndexOf(IItem);
            IItem:=TItemInfo(FItems[FItems.Count-1]);
            for t:=LWorm.FItems.Count-1 downto c do begin
              LItem:=TItemInfo(LWorm.FItems[t]);
              LItem.FX:=IItem.FX;
              LItem.FY:=IItem.FY;
              LItem.FIX:=IItem.FIX;
              LItem.FIY:=IItem.FIY;
              LItem.FLevel:=IItem.FLevel;
              LItem.FOwner:=IItem.FOwner;
              FItems.Add(LItem);
              LWorm.FItems.Delete(t);
            end;
            if LWorm.FItems.Count=0 then begin
              Self.FWorms.Delete(Self.FWorms.IndexOf(LWorm));
              LWorm.Free;
            end;
            break;
          end;
    end;

    if FNextStep then FNextStep:=not (((FNX<0) or (FNX>FVDX)) or ((FNY<0) or (FNY>FVDY)));

    if FNextStep then begin

      FSkip:=0;

      LItem:=TItemInfo(FItems[FItems.Count-1]);
      for i:=FItems.Count-2 downto 0 do begin
        IItem:=TItemInfo(FItems[i]);
        LItem.FX:=IItem.FX;
        LItem.FY:=IItem.FY;
        if (LItem.FIX<>IItem.FIX) or (LItem.FIY<>IItem.FIY)
        then LItem.FR:=1
        else LItem.FR:=0;

        LItem.FIX:=IItem.FIX;
        LItem.FIY:=IItem.FIY;

        LItem.FLevel:=IItem.FLevel;
        LItem.FStep:=FWormStep;
        LItem:=IItem;
      end;

      LItem.FX:=FNX*FItemSpacing;
      LItem.FY:=FNY*FItemSpacing;
      LItem.FIX:=FNX;
      LItem.FIY:=FNY;
      LItem.FR:=1;
      LItem.FLevel:=FLevel;
      LItem.FStep:=FWormStep;

      Dec(FStep);
    end else begin
      FStep:=0;
      inc(FSkip);
    end;


  end;

end;

function DeskIconThread.FreeAtoms(AStoped: boolean): integer;
var
  i : integer;
begin
  result:=100;
  FOnProcess:=ProcessWait;
  for i:=0 to FItems.Count-1 do TItemInfo(FItems[i]).Free;
  FItems.Free;
end;

function DeskIconThread.InitAtoms(AStoped: boolean): integer;
var
  p : TPoint;
  i,c,n : integer;
  LItem : TItemInfo;
  LItemRect : TRect;
begin
  result:=10;
  FOnProcess:=ProcessWait;
  if AStoped then exit;

  c:=ListView_GetItemCount(HDeskList);
  if c=0 then exit;
  if not ListView_GetViewRect(HDeskList,FViewPort) then exit;
  if not ListView_GetItemRect(HDeskList,0,LItemRect,LVIR_BOUNDS) then exit;

  FItemSpacing:=((LItemRect.Right-LItemRect.Left)+(LItemRect.Bottom-LItemRect.Top)) div 2;

  FItems := TList.Create;
  FOnProcess:=ProcessAtoms;

  FAtomStep :=0;

  for i:=0 to c-1 do begin
    if ListView_GetItemPosition(HDeskList, i, p) then begin
      LItem := TItemInfo.Create;
      LItem.FX:= p.x;
      LItem.FY:= p.y;
      LItem.FDX:= random(10)-5;
      LItem.FDY:= random(10)-5;
      LItem.FIndex:=i;
      LItem.FR := 180+random(40);
      FItems.Add(LItem);
    end else break;
  end;

  for i:=0 to FItems.Count-1 do begin
    LItem:=TItemInfo(FItems[i]);
    n:=i+1;
    if n=FItems.Count then n:=0;
    LItem.FLinks.Add(FItems[n]);
    TItemInfo(FItems[n]).FLinks.Add(LItem);
  end;

end;

function DeskIconThread.ProcessAtoms(AStoped: boolean): integer;
var
  i,t : integer;
  LItem,IItem : TItemInfo;
  dx,dy,vx,vy,len,dlen : extended;

  procedure moveone(pn,pc:TItemInfo);
  begin
    PN.FDX:=PN.FDX+((PC.FX-PN.FX))/PN.FR;
    PN.FDY:=PN.FDY+((PC.FY-PN.FY))/PN.FR;
  end;

begin
  result:=100;
  FOnProcess:=FreeAtoms;
  if AStoped then exit;
  Inc(FAtomStep);

  if (FAtomStep<ATOMSLIFECYCLE) then FOnProcess:=ProcessAtoms;

  for i:=0 to FItems.Count-1 do begin
    LItem:=TItemInfo(FItems[i]);
    for t:=0 to LItem.FLinks.Count-1 do
      moveone(LItem,TItemInfo(LItem.FLinks[t]));
  end;

  for i:=0 to FItems.Count-1 do begin
    LItem:=TItemInfo(FItems[i]);
    dx := 0;
    dy := 0;
    for t:=0 to FItems.Count-1 do begin
      if i=t then continue;
      IItem:=TItemInfo(FItems[t]);
      vx := LItem.FX-IItem.FX;
      vy := LItem.FY-IItem.FY;
      len:=vx * vx + vy * vy;
      if len=0 then begin
        dx := dx+random;
        dy := dy+random;
      end else if len<((FItemSpacing+FItemSpacing)*(FItemSpacing+FItemSpacing)) then begin
        dx := dx+vx/len;
        dy := dy+vy/len;
      end;
    end;
    dlen:=dx * dx + dy * dy;
    if dlen>0 then begin
      dlen := sqrt(dlen) / 2;
      LItem.FDX := LItem.FDX+dx / dlen;
      LItem.FDY := LItem.FDY+dy / dlen;
    end;
  end;

  for i:=0 to FItems.Count-1 do begin
    LItem:=TItemInfo(FItems[i]);

    if (LItem.FX+LItem.FDX) > FViewPort.Right then begin
      LItem.FDX:=FViewPort.Right-(LItem.FX+LItem.FDX);
      LItem.FX:=FViewPort.Right;
    end else LItem.FX:=LItem.FX+LItem.FDX;

    if (LItem.FX+LItem.FDX) < FViewPort.Left then begin
      LItem.FDX:=FViewPort.Left-(LItem.FX+LItem.FDX);
      LItem.FX:=FViewPort.Left;
    end else LItem.FX:=LItem.FX+LItem.FDX;

    if (LItem.FY+LItem.FDY) > FViewPort.Bottom then begin
      LItem.FDY:=FViewPort.Bottom-(LItem.FY+LItem.FDY);
      LItem.FY:=FViewPort.Bottom;
    end else LItem.FY:=LItem.FY+LItem.FDY;

    if (LItem.FY+LItem.FDY) < FViewPort.Top then begin
      LItem.FDY:=FViewPort.Top-(LItem.FY+LItem.FDY);
      LItem.FY:=FViewPort.Top;
    end else LItem.FY:=LItem.FY+LItem.FDY;

    ListView_SetItemPosition(HDeskList,LItem.FIndex,trunc(LItem.FX),trunc(LItem.FY));
  end;

end;

function DeskIconThread.FreeMouse(AStoped: boolean): integer;
var
  i : integer;
begin
  result:=100;
  FOnProcess:=ProcessWait;
  for i:=0 to FItems.Count-1 do TItemInfo(FItems[i]).Free;
  FItems.Free;
end;

function DeskIconThread.InitMouse(AStoped: boolean): integer;
var
  i,c : integer;
  LItem : TItemInfo;
  LItemRect : TRect;
begin
  result:=10;
  FOnProcess:=ProcessWait;
  if AStoped then exit;

  c:=ListView_GetItemCount(HDeskList);
  if c=0 then exit;
  if not ListView_GetViewRect(HDeskList,FViewPort) then exit;
  if not ListView_GetItemRect(HDeskList,0,LItemRect,LVIR_BOUNDS) then exit;

  FItemSpacing:=((LItemRect.Right-LItemRect.Left)+(LItemRect.Bottom-LItemRect.Top)) div 2;
  InflateRect(FViewPort,-(FItemSpacing div 2),-(FItemSpacing div 2));
  FItems := TList.Create;
  FOnProcess:=ProcessMouse;

  FAtomStep :=0;

  for i:=0 to c-1 do begin
      LItem := TItemInfo.Create;
      LItem.FDX:= 0;
      LItem.FDY:= 0;
      LItem.FIndex:=i;
      LItem.FR := random(20)+5;
      FItems.Add(LItem);
  end;

end;

function DeskIconThread.ProcessMouse(AStoped: boolean): integer;
var
  p : TPoint;
  i,t : integer;
  LItem,IItem : TItemInfo;
  dx,dy,vx,vy,len,dlen : extended;

begin
  result:=100;
  FOnProcess:=FreeMouse;
  if AStoped then exit;
  Inc(FAtomStep);

  if (FAtomStep<MOUSESLIFECYCLE) then FOnProcess:=ProcessMouse;

  if not GetCursorPos(FMouseP) then exit;

  for i:=0 to FItems.Count-1 do begin
    LItem:=TItemInfo(FItems[i]);
    if ListView_GetItemPosition(HDeskList, LItem.FIndex, p) then begin
      LItem.FX:= p.x+(FItemSpacing div 2);
      LItem.FY:= p.y+(FItemSpacing div 2);
    end;
  end;

  for i:=0 to FItems.Count-1 do begin
    LItem:=TItemInfo(FItems[i]);
    dx := 0;
    dy := 0;
    for t:=0 to FItems.Count-1 do begin
      IItem:=TItemInfo(FItems[t]);
      if i=t then begin
        vx := LItem.FX-FMouseP.X;
        vy := LItem.FY-FMouseP.Y;
      end else begin
        vx := LItem.FX-IItem.FX;
        vy := LItem.FY-IItem.FY;
      end;
      len:=vx * vx + vy * vy;
      if len=0 then begin
        dx := dx+random;
        dy := dy+random;
      end else if len<((FItemSpacing+FItemSpacing div 10)*(FItemSpacing+FItemSpacing div 10)) then begin
        dx := dx+vx/len;
        dy := dy+vy/len;
      end;
    end;
    dlen:=dx * dx + dy * dy;
    if (dlen>0) then begin
      dlen := sqrt(dlen) / 2;
      LItem.FDX := LItem.FDX+dx / dlen;
      LItem.FDY := LItem.FDY+dy / dlen;
    end else begin
      LItem.FDX := 0;
      LItem.FDY := 0;
    end;
  end;

  for i:=0 to FItems.Count-1 do begin
    LItem:=TItemInfo(FItems[i]);

    if (LItem.FDX=0) and (LItem.FDY=0) then continue
    else begin
      LItem.FCDX:=max(-LItem.FR, min(LItem.FR, LItem.FDX));
      LItem.FCDY:=max(-LItem.FR, min(LItem.FR, LItem.FDY));
    end;

    if (LItem.FX+LItem.FCDX) > FViewPort.Right then begin
      LItem.FDX:=FViewPort.Right-(LItem.FX+LItem.FCDX);
      LItem.FX:=FViewPort.Right;
    end else LItem.FX:=LItem.FX+LItem.FCDX;

    if (LItem.FX+LItem.FCDX) < FViewPort.Left then begin
      LItem.FDX:=FViewPort.Left-(LItem.FX+LItem.FCDX);
      LItem.FX:=FViewPort.Left;
    end else LItem.FX:=LItem.FX+LItem.FCDX;

    if (LItem.FY+LItem.FCDY) > FViewPort.Bottom then begin
      LItem.FDY:=FViewPort.Bottom-(LItem.FY+LItem.FCDY);
      LItem.FY:=FViewPort.Bottom;
    end else LItem.FY:=LItem.FY+LItem.FCDY;

    if (LItem.FY+LItem.FCDY) < FViewPort.Top then begin
      LItem.FDY:=FViewPort.Top-(LItem.FY+LItem.FCDY);
      LItem.FY:=FViewPort.Top;
    end else LItem.FY:=LItem.FY+LItem.FCDY;

    ListView_SetItemPosition(HDeskList,LItem.FIndex,trunc(LItem.FX-(FItemSpacing div 2)),trunc(LItem.FY-(FItemSpacing div 2)));
  end;

end;
{ TWormInfo }

constructor TWormInfo.Create;
begin
  inherited Create;
  FItems := TList.Create;
  FStep:=0;
end;

destructor TWormInfo.Destroy;
begin
  FItems.Free;
  inherited Destroy;
end;

{ TItemInfo }

constructor TItemInfo.Create;
begin
  inherited Create;
  FLinks := TList.Create;
  FStep:=0;
end;

destructor TItemInfo.Destroy;
begin
  FLinks.Free;
  inherited Destroy;
end;

end.
